@@ -20,6 +20,7 @@ VALUE rb_cArray;
 static ID id_cmp;
 
 #define ARY_DEFAULT_SIZE 16
+#define ARY_MAX_SIZE (LONG_MAX / sizeof(VALUE))
 
 void
 rb_mem_clear(register VALUE *mem, register long size)
@@ -114,7 +115,7 @@ ary_new(VALUE klass, long len)
     if (len < 0) {
 	rb_raise(rb_eArgError, "negative array size (or size too big)");
     }
-    if (len > 0 && len * sizeof(VALUE) <= len) {
+    if (len > ARY_MAX_SIZE) {
 	rb_raise(rb_eArgError, "array size too big");
     }
     ary = ary_alloc(klass);
@@ -290,9 +291,9 @@ rb_ary_initialize(int argc, VALUE *argv, VALUE ary)
     VALUE size, val;
 
     rb_ary_modify(ary);
-    if (rb_scan_args(argc, argv, "02", &size, &val) == 0) {
+    if (argc ==  0) {
 	if (RARRAY_PTR(ary) && !ARY_SHARED_P(ary)) {
-	    free(RARRAY(ary)->ptr);
+	    xfree(RARRAY(ary)->ptr);
 	}
 	RARRAY(ary)->len = 0;
 	if (rb_block_given_p()) {
@@ -300,7 +301,7 @@ rb_ary_initialize(int argc, VALUE *argv, VALUE ary)
 	}
 	return ary;
     }
-
+    rb_scan_args(argc, argv, "02", &size, &val);
     if (argc == 1 && !FIXNUM_P(size)) {
 	val = rb_check_array_type(size);
 	if (!NIL_P(val)) {
@@ -313,7 +314,7 @@ rb_ary_initialize(int argc, VALUE *argv, VALUE ary)
     if (len < 0) {
 	rb_raise(rb_eArgError, "negative array size");
     }
-    if (len > 0 && len * (long)sizeof(VALUE) <= len) {
+    if (len > ARY_MAX_SIZE) {
 	rb_raise(rb_eArgError, "array size too big");
     }
     rb_ary_modify(ary);
@@ -371,6 +372,9 @@ rb_ary_store(VALUE ary, long idx, VALUE val)
 		     idx - RARRAY_LEN(ary));
 	}
     }
+    else if (idx >= ARY_MAX_SIZE) {
+	rb_raise(rb_eIndexError, "index %ld too big", idx);
+    }
 
     rb_ary_modify(ary);
     if (idx >= ARY_CAPA(ary)) {
@@ -379,13 +383,10 @@ rb_ary_store(VALUE ary, long idx, VALUE val)
 	if (new_capa < ARY_DEFAULT_SIZE) {
 	    new_capa = ARY_DEFAULT_SIZE;
 	}
-	if (new_capa + idx < new_capa) {
-	    rb_raise(rb_eArgError, "index too big");
+	if (new_capa >= ARY_MAX_SIZE - idx) {
+	    new_capa = (ARY_MAX_SIZE - idx) / 2;
 	}
 	new_capa += idx;
-	if (new_capa * (long)sizeof(VALUE) <= new_capa) {
-	    rb_raise(rb_eArgError, "index too big");
-	}
 	RESIZE_CAPA(ary, new_capa);
     }
     if (idx > RARRAY_LEN(ary)) {
@@ -498,10 +499,14 @@ rb_ary_pop(VALUE ary)
 
 /*
  *  call-seq:
- *     array.pop  -> obj or nil
+ *     array.pop    -> obj or nil
+ *     array.pop(n) -> array
  *  
  *  Removes the last element from <i>self</i> and returns it, or
  *  <code>nil</code> if the array is empty.
+ *
+ *  If a number _n_ is given, returns an array of the last n elements
+ *  (or less) just like <code>array.slice!(-n, n)</code> does.
  *     
  *     a = [ "a", "b", "c", "d" ]
  *     a.pop     #=> "d"
@@ -549,11 +554,15 @@ rb_ary_shift(VALUE ary)
 
 /*
  *  call-seq:
- *     array.shift   ->   obj or nil
+ *     array.shift    -> obj or nil
+ *     array.shift(n) -> array
  *  
  *  Returns the first element of <i>self</i> and removes it (shifting all
  *  other elements down by one). Returns <code>nil</code> if the array
  *  is empty.
+ *
+ *  If a number _n_ is given, returns an array of the first n elements
+ *  (or less) just like <code>array.slice!(0, n)</code> does.
  *     
  *     args = [ "-m", "-q", "filename" ]
  *     args.shift     #=> "-m"
@@ -869,6 +878,8 @@ rb_ary_fetch(int argc, VALUE *argv, VALUE ary)
  *     a.index("b")        #=> 1
  *     a.index("z")        #=> nil
  *     a.index{|x|x=="b"}  #=> 1
+ *
+ *  This is an alias of <code>#find_index</code>.
  */
 
 static VALUE
@@ -877,19 +888,19 @@ rb_ary_index(int argc, VALUE *argv, VALUE ary)
     VALUE val;
     long i;
 
-    if (rb_scan_args(argc, argv, "01", &val) == 0) {
+    if (argc  == 0) {
 	RETURN_ENUMERATOR(ary, 0, 0);
 	for (i=0; i<RARRAY_LEN(ary); i++) {
 	    if (RTEST(rb_yield(RARRAY_PTR(ary)[i]))) {
 		return LONG2NUM(i);
 	    }
 	}
+	return Qnil;
     }
-    else {
-	for (i=0; i<RARRAY_LEN(ary); i++) {
-	    if (rb_equal(RARRAY_PTR(ary)[i], val))
-		return LONG2NUM(i);
-	}
+    rb_scan_args(argc, argv, "01", &val);
+    for (i=0; i<RARRAY_LEN(ary); i++) {
+	if (rb_equal(RARRAY_PTR(ary)[i], val))
+	    return LONG2NUM(i);
     }
     return Qnil;
 }
@@ -915,7 +926,7 @@ rb_ary_rindex(int argc, VALUE *argv, VALUE ary)
     VALUE val;
     long i = RARRAY_LEN(ary);
 
-    if (rb_scan_args(argc, argv, "01", &val) == 0) {
+    if (argc == 0) {
 	RETURN_ENUMERATOR(ary, 0, 0);
 	while (i--) {
 	    if (RTEST(rb_yield(RARRAY_PTR(ary)[i])))
@@ -924,14 +935,14 @@ rb_ary_rindex(int argc, VALUE *argv, VALUE ary)
 		i = RARRAY_LEN(ary);
 	    }
 	}
+	return Qnil;
     }
-    else {
-	while (i--) {
-	    if (rb_equal(RARRAY_PTR(ary)[i], val))
-		return LONG2NUM(i);
-	    if (i > RARRAY_LEN(ary)) {
-		i = RARRAY_LEN(ary);
-	    }
+    rb_scan_args(argc, argv, "01", &val);
+    while (i--) {
+	if (rb_equal(RARRAY_PTR(ary)[i], val))
+	    return LONG2NUM(i);
+	if (i > RARRAY_LEN(ary)) {
+	    i = RARRAY_LEN(ary);
 	}
     }
     return Qnil;
@@ -975,6 +986,9 @@ rb_ary_splice(VALUE ary, long beg, long len, VALUE rpl)
     }
     rb_ary_modify(ary);
     if (beg >= RARRAY_LEN(ary)) {
+	if (beg > ARY_MAX_SIZE - rlen) {
+	    rb_raise(rb_eIndexError, "index %ld too big", beg);
+	}
 	len = beg + rlen;
 	if (len >= ARY_CAPA(ary)) {
 	    RESIZE_CAPA(ary, len);
@@ -1033,7 +1047,7 @@ rb_ary_splice(VALUE ary, long beg, long len, VALUE rpl)
  *     a[0..2] = "A"               #=> ["A", "4"]
  *     a[-1]   = "Z"               #=> ["A", "Z"]
  *     a[1..-1] = nil              #=> ["A", nil]
- *     a[1..-1] = []              #=> ["A"]
+ *     a[1..-1] = []               #=> ["A"]
  */
 
 static VALUE
@@ -1219,9 +1233,9 @@ rb_ary_dup(VALUE ary)
 {
     VALUE dup = rb_ary_new2(RARRAY_LEN(ary));
 
+    DUPSETUP(dup, ary);
     MEMCPY(RARRAY_PTR(dup), RARRAY_PTR(ary), VALUE, RARRAY_LEN(ary));
     RARRAY(dup)->len = RARRAY_LEN(ary);
-    OBJ_INFECT(dup, ary);
 
     return dup;
 }
@@ -1432,21 +1446,32 @@ rb_ary_reverse_m(VALUE ary)
     return rb_ary_reverse(rb_ary_dup(ary));
 }
 
+static VALUE
+sort_reentered(VALUE *klass)
+{
+    if (*klass) {
+	rb_raise(rb_eRuntimeError, "sort reentered");
+    }
+    return Qnil;
+}
+
 static int
 sort_1(const void *ap, const void *bp, void *dummy)
 {
+    VALUE retval = sort_reentered(dummy);
     VALUE a = *(const VALUE *)ap, b = *(const VALUE *)bp;
-    VALUE retval = rb_yield_values(2, a, b);
     int n;
 
+    retval = rb_yield_values(2, a, b);
     n = rb_cmpint(retval, a, b);
+    sort_reentered(dummy);
     return n;
 }
 
 static int
 sort_2(const void *ap, const void *bp, void *dummy)
 {
-    VALUE retval;
+    VALUE retval = sort_reentered(dummy);
     VALUE a = *(const VALUE *)ap, b = *(const VALUE *)bp;
     int n;
 
@@ -1461,6 +1486,7 @@ sort_2(const void *ap, const void *bp, void *dummy)
 
     retval = rb_funcall(a, id_cmp, 1, b);
     n = rb_cmpint(retval, a, b);
+    sort_reentered(dummy);
 
     return n;
 }
@@ -1490,12 +1516,19 @@ rb_ary_sort_bang(VALUE ary)
 
 	RBASIC(tmp)->klass = 0;
 	ruby_qsort(RARRAY_PTR(tmp), RARRAY_LEN(tmp), sizeof(VALUE),
-		   rb_block_given_p()?sort_1:sort_2, 0);
-	RARRAY(ary)->ptr = RARRAY(tmp)->ptr;
-	RARRAY(ary)->len = RARRAY(tmp)->len;
-	RARRAY(ary)->aux.capa = RARRAY(tmp)->aux.capa;
+		   rb_block_given_p()?sort_1:sort_2, &RBASIC(tmp)->klass);
+	if (RARRAY(ary)->ptr != RARRAY(tmp)->ptr) {
+	    if (!ARY_SHARED_P(ary)) xfree(RARRAY(ary)->ptr);
+	    RARRAY(ary)->ptr = RARRAY(tmp)->ptr;
+	    RARRAY(ary)->len = RARRAY(tmp)->len;
+	    RARRAY(ary)->aux.capa = RARRAY(tmp)->aux.capa;
+	    FL_SET(ary, ELTS_SHARED);
+	};
 	FL_UNSET(ary, ELTS_SHARED);
-	rb_gc_force_recycle(tmp);
+	RARRAY(tmp)->ptr = 0;
+	RARRAY(tmp)->len = 0;
+	RARRAY(tmp)->aux.capa = 0;
+	RBASIC(tmp)->klass = RBASIC(ary)->klass;
     }
     return ary;
 }
@@ -1759,13 +1792,7 @@ rb_ary_delete_at_m(VALUE ary, VALUE pos)
  *  
  *  Deletes the element(s) given by an index (optionally with a length)
  *  or by a range. Returns the deleted object, subarray, or
- *  <code>nil</code> if the index is out of range. Equivalent to:
- *     
- *     def slice!(*args)
- *       result = self[*args]
- *       self[*args] = nil
- *       result
- *     end
+ *  <code>nil</code> if the index is out of range.
  *     
  *     a = [ "a", "b", "c" ]
  *     a.slice!(1)     #=> "b"
@@ -1780,18 +1807,27 @@ static VALUE
 rb_ary_slice_bang(int argc, VALUE *argv, VALUE ary)
 {
     VALUE arg1, arg2;
-    long pos, len;
+    long pos, len, orig_len;
 
+    rb_ary_modify_check(ary);
     if (rb_scan_args(argc, argv, "11", &arg1, &arg2) == 2) {
 	pos = NUM2LONG(arg1);
 	len = NUM2LONG(arg2);
       delete_pos_len:
+	if (len < 0) return Qnil;
+	orig_len = RARRAY_LEN(ary);
 	if (pos < 0) {
-	    pos = RARRAY_LEN(ary) + pos;
+	    pos += orig_len;
 	    if (pos < 0) return Qnil;
 	}
-	arg2 = rb_ary_subseq(ary, pos, len);
-	rb_ary_splice(ary, pos, len, Qundef);	/* Qnil/rb_ary_new2(0) */
+	else if (orig_len < pos) return Qnil;
+	if (orig_len < pos + len) {
+	    len = orig_len - pos;
+	}
+	if (len == 0) return rb_ary_new2(0);
+	arg2 = rb_ary_new4(len, RARRAY_PTR(ary)+pos);
+	RBASIC(arg2)->klass = rb_obj_class(ary);
+	rb_ary_splice(ary, pos, len, Qundef);
 	return arg2;
     }
 
@@ -1875,6 +1911,7 @@ rb_ary_reject(VALUE ary)
 static VALUE
 rb_ary_delete_if(VALUE ary)
 {
+    RETURN_ENUMERATOR(ary, 0, 0);
     rb_ary_reject_bang(ary);
     return ary;
 }
@@ -1916,17 +1953,13 @@ take_items(VALUE obj, long n)
  *     
  *     a = [ 4, 5, 6 ]
  *     b = [ 7, 8, 9 ]
- *     
  *     [1,2,3].zip(a, b)      #=> [[1, 4, 7], [2, 5, 8], [3, 6, 9]]
  *     [1,2].zip(a,b)         #=> [[1, 4, 7], [2, 5, 8]]
  *     a.zip([1,2],[8])       #=> [[4,1,8], [5,2,nil], [6,nil,nil]]
  */
 
 static VALUE
-rb_ary_zip(argc, argv, ary)
-    int argc;
-    VALUE *argv;
-    VALUE ary;
+rb_ary_zip(int argc, VALUE *argv, VALUE ary)
 {
     int i, j;
     long len;
@@ -2221,7 +2254,7 @@ rb_ary_times(VALUE ary, VALUE times)
     if (len < 0) {
 	rb_raise(rb_eArgError, "negative argument");
     }
-    if (LONG_MAX/len < RARRAY_LEN(ary)) {
+    if (ARY_MAX_SIZE/len < RARRAY_LEN(ary)) {
 	rb_raise(rb_eArgError, "argument too big");
     }
     len *= RARRAY_LEN(ary);
@@ -2684,11 +2717,13 @@ rb_ary_compact_bang(VALUE ary)
 	if (NIL_P(*t)) t++;
 	else *p++ = *t++;
     }
-    if (RARRAY_LEN(ary) == (p - RARRAY_PTR(ary))) {
+    n = p - RARRAY_PTR(ary);
+    if (RARRAY_LEN(ary) == n) {
 	return Qnil;
     }
-    n = p - RARRAY_PTR(ary);
-    RESIZE_CAPA(ary, n);
+    if (n * 2 < ARY_CAPA(ary) && ARY_DEFAULT_SIZE * 2 < ARY_CAPA(ary)) {
+	RESIZE_CAPA(ary, n * 2);
+    }
     RARRAY(ary)->len = n;
 
     return ary;
@@ -2714,41 +2749,48 @@ rb_ary_compact(VALUE ary)
 
 /*
  *  call-seq:
- *     array.nitems -> int
- *     array.nitems { |item| block }  -> int
+ *     array.count      -> int
+ *     array.count(obj) -> int
+ *     array.count { |item| block }  -> int
  *  
- *  Returns the number of non-<code>nil</code> elements in _self_.
- *  If a block is given, the elements yielding a true value are
- *  counted.
+ *  Returns the number of elements.  If an argument is given, counts
+ *  the number of elements which equals to <i>obj</i>.  If a block is
+ *  given, counts the number of elements yielding a true value.
+ *
+ *     ary = [1, 2, 4, 2]
+ *     ary.count             # => 4
+ *     ary.count(2)          # => 2
+ *     ary.count{|x|x%2==0}  # => 3
  *
- *  May be zero.
- *     
- *     [ 1, nil, 3, nil, 5 ].nitems   #=> 3
- *     [5,6,7,8,9].nitems { |x| x % 2 != 0 }  #=> 3
  */
 
 static VALUE
-rb_ary_nitems(VALUE ary)
+rb_ary_count(int argc, VALUE *argv, VALUE ary)
 {
     long n = 0;
- 
-    if (rb_block_given_p()) {
-	long i;
 
-	for (i=0; i<RARRAY_LEN(ary); i++) {
-	    VALUE v = RARRAY_PTR(ary)[i];
- 	    if (RTEST(rb_yield(v))) n++;
- 	}
+    if (argc == 0) {
+	VALUE *p, *pend;
+
+	if (!rb_block_given_p())
+	    return LONG2NUM(RARRAY_LEN(ary));
+
+	for (p = RARRAY_PTR(ary), pend = p + RARRAY_LEN(ary); p < pend; p++) {
+	    if (RTEST(rb_yield(*p))) n++;
+	}
     }
     else {
-	VALUE *p = RARRAY_PTR(ary);
-	VALUE *pend = p + RARRAY_LEN(ary);
+	VALUE obj, *p, *pend;
 
- 	while (p < pend) {
- 	    if (!NIL_P(*p)) n++;
- 	    p++;
- 	}
+	rb_scan_args(argc, argv, "1", &obj);
+	if (rb_block_given_p()) {
+	    rb_warn("given block not used");
+	}
+	for (p = RARRAY_PTR(ary), pend = p + RARRAY_LEN(ary); p < pend; p++) {
+	    if (rb_equal(*p, obj)) n++;
+	}
     }
+
     return LONG2NUM(n);
 }
 
@@ -2760,8 +2802,8 @@ flatten(VALUE ary, int level, int *modified)
     st_table *memo;
     st_data_t id;
 
-    stack = rb_ary_new();
-    result = rb_ary_new();
+    stack = ary_new(0, ARY_DEFAULT_SIZE);
+    result = ary_new(0, RARRAY_LEN(ary));
     memo = st_init_numtable();
     st_insert(memo, (st_data_t)ary, (st_data_t)Qtrue);
     *modified = 0;
@@ -2770,6 +2812,9 @@ flatten(VALUE ary, int level, int *modified)
 	while (i < RARRAY_LEN(ary)) {
 	    elt = RARRAY_PTR(ary)[i++];
 	    tmp = rb_check_array_type(elt);
+	    if (RBASIC(result)->klass) {
+		rb_raise(rb_eRuntimeError, "flatten reentered");
+	    }
 	    if (NIL_P(tmp) || (level >= 0 && RARRAY_LEN(stack) / 2 >= level)) {
 		rb_ary_push(result, elt);
 	    }
@@ -2777,6 +2822,7 @@ flatten(VALUE ary, int level, int *modified)
 		*modified = 1;
 		id = (st_data_t)tmp;
 		if (st_lookup(memo, id, 0)) {
+		    st_free_table(memo);
 		    rb_raise(rb_eArgError, "tried to flatten recursive array");
 		}
 		st_insert(memo, id, (st_data_t)Qtrue);
@@ -2796,6 +2842,9 @@ flatten(VALUE ary, int level, int *modified)
 	ary = rb_ary_pop(stack);
     }
 
+    st_free_table(memo);
+
+    RBASIC(result)->klass = rb_class_of(ary);
     return result;
 }
 
@@ -2847,7 +2896,7 @@ rb_ary_flatten_bang(int argc, VALUE *argv, VALUE ary)
  *     s = [ 1, 2, 3 ]           #=> [1, 2, 3]
  *     t = [ 4, 5, 6, [7, 8] ]   #=> [4, 5, 6, [7, 8]]
  *     a = [ s, t, 9, 10 ]       #=> [[1, 2, 3], [4, 5, 6, [7, 8]], 9, 10]
- *     a.flatten                 #=> [1, 2, 3, 4, 5, 6, 7, 8, 9, 10
+ *     a.flatten                 #=> [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
  *     a = [ 1, 2, [3, [4, 5] ] ]
  *     a.flatten(1)              #=> [1, 2, 3, [4, 5]]
  */
@@ -2870,7 +2919,7 @@ rb_ary_flatten(int argc, VALUE *argv, VALUE ary)
 
 /*
  *  call-seq:
- *     array.shuffle!        -> array or nil
+ *     array.shuffle!        -> array
  *  
  *  Shuffles elements in _self_ in place.
  */
@@ -2934,24 +2983,40 @@ rb_ary_choice(VALUE ary)
 /*
  *  call-seq:
  *     ary.cycle {|obj| block }
+ *     ary.cycle(n) {|obj| block }
  *  
- *  Calls <i>block</i> repeatedly forever.
+ *  Calls <i>block</i> for each element repeatedly _n_ times or
+ *  forever if none or nil is given.  If a non-positive number is
+ *  given or the array is empty, does nothing.  Returns nil if the
+ *  loop has finished without getting interrupted.
  *     
  *     a = ["a", "b", "c"]
  *     a.cycle {|x| puts x }  # print, a, b, c, a, b, c,.. forever.
+ *     a.cycle(2) {|x| puts x }  # print, a, b, c, a, b, c.
  *     
  */
 
 static VALUE
-rb_ary_cycle(VALUE ary)
+rb_ary_cycle(int argc, VALUE *argv, VALUE ary)
 {
-    long i;
+    long n, i;
+    VALUE nv = Qnil;
 
-    RETURN_ENUMERATOR(ary, 0, 0);
-    while (RARRAY_LEN(ary) > 0) {
-	for (i=0; i<RARRAY_LEN(ary); i++) {
-	    rb_yield(RARRAY_PTR(ary)[i]);
-	}
+    rb_scan_args(argc, argv, "01", &nv);
+
+    RETURN_ENUMERATOR(ary, argc, argv);
+    if (NIL_P(nv)) {
+        n = -1;
+    }
+    else {
+        n = NUM2LONG(nv);
+        if (n <= 0) return Qnil;
+    }
+
+    while (RARRAY_LEN(ary) > 0 && (n < 0 || 0 < n--)) {
+        for (i=0; i<RARRAY_LEN(ary); i++) {
+            rb_yield(RARRAY_PTR(ary)[i]);
+        }
     }
     return Qnil;
 }
@@ -3015,13 +3080,14 @@ permute0(long n, long r, long *p, long index, int *used, VALUE values)
  * When invoked without a block, return an enumerator object instead.
  * 
  * Examples:
+ *
  *     a = [1, 2, 3]
  *     a.permutation.to_a     #=> [[1,2,3],[1,3,2],[2,1,3],[2,3,1],[3,1,2],[3,2,1]]
  *     a.permutation(1).to_a  #=> [[1],[2],[3]]
  *     a.permutation(2).to_a  #=> [[1,2],[1,3],[2,1],[2,3],[3,1],[3,2]]
  *     a.permutation(3).to_a  #=> [[1,2,3],[1,3,2],[2,1,3],[2,3,1],[3,1,2],[3,2,1]]
- *     a.permutation(0).to_a  #=> [[]]: one permutation of length 0
- *     a.permutation(4).to_a  #=> []  : no permutations of length 4
+ *     a.permutation(0).to_a  #=> [[]] # one permutation of length 0
+ *     a.permutation(4).to_a  #=> []   # no permutations of length 4
  */
 
 static VALUE
@@ -3030,8 +3096,8 @@ rb_ary_permutation(int argc, VALUE *argv, VALUE ary)
     VALUE num;
     long r, n, i;
 
-    RETURN_ENUMERATOR(ary, argc, argv);   /* Return enumerator if no block */
     n = RARRAY_LEN(ary);                  /* Array length */
+    RETURN_ENUMERATOR(ary, argc, argv);   /* Return enumerator if no block */
     rb_scan_args(argc, argv, "01", &num);
     r = NIL_P(num) ? n : NUM2LONG(num);   /* Permutation size from argument */
 
@@ -3095,13 +3161,14 @@ combi_len(long n, long k)
  * When invoked without a block, returns an enumerator object instead.
  *     
  * Examples:
+ *
  *     a = [1, 2, 3, 4]
  *     a.combination(1).to_a  #=> [[1],[2],[3],[4]]
  *     a.combination(2).to_a  #=> [[1,2],[1,3],[1,4],[2,3],[2,4],[3,4]]
  *     a.combination(3).to_a  #=> [[1,2,3],[1,2,4],[1,3,4],[2,3,4]]
  *     a.combination(4).to_a  #=> [[1,2,3,4]]
- *     a.combination(0).to_a  #=> [[]]: one combination of length 0
- *     a.combination(5).to_a  #=> []  : no combinations of length 5
+ *     a.combination(0).to_a  #=> [[]] # one combination of length 0
+ *     a.combination(5).to_a  #=> []   # no combinations of length 5
  *     
  */
 
@@ -3110,8 +3177,8 @@ rb_ary_combination(VALUE ary, VALUE num)
 {
     long n, i, len;
 
-    RETURN_ENUMERATOR(ary, 1, &num);
     n = NUM2LONG(num);
+    RETURN_ENUMERATOR(ary, 1, &num);
     len = RARRAY_LEN(ary);
     if (n < 0 || len < n) {
 	/* yield nothing */
@@ -3226,6 +3293,102 @@ rb_ary_product(int argc, VALUE *argv, VALUE ary)
     return result;
 }
 
+/*
+ *  call-seq:
+ *     ary.take(n)               => array
+ *  
+ *  Returns first n elements from <i>ary</i>.
+ *     
+ *     a = [1, 2, 3, 4, 5, 0]
+ *     a.take(3)             # => [1, 2, 3]
+ *     
+ */
+
+static VALUE
+rb_ary_take(VALUE obj, VALUE n)
+{
+    long len = NUM2LONG(n);
+    if (len < 0) {
+	rb_raise(rb_eArgError, "attempt to take negative size");
+    }
+    return rb_ary_subseq(obj, 0, len);
+}
+
+/*
+ *  call-seq:
+ *     ary.take_while {|arr| block }   => array
+ *  
+ *  Passes elements to the block until the block returns nil or false,
+ *  then stops iterating and returns an array of all prior elements.
+ *     
+ *     a = [1, 2, 3, 4, 5, 0]
+ *     a.take_while {|i| i < 3 }   # => [1, 2]
+ *     
+ */
+
+static VALUE
+rb_ary_take_while(VALUE ary)
+{
+    long i;
+
+    RETURN_ENUMERATOR(ary, 0, 0);
+    for (i = 0; i < RARRAY_LEN(ary); i++) {
+	if (!RTEST(rb_yield(RARRAY_PTR(ary)[i]))) break;
+    }
+    return rb_ary_take(ary, LONG2FIX(i));
+}
+
+/*
+ *  call-seq:
+ *     ary.drop(n)               => array
+ *  
+ *  Drops first n elements from <i>ary</i>, and returns rest elements
+ *  in an array.
+ *     
+ *     a = [1, 2, 3, 4, 5, 0]
+ *     a.drop(3)             # => [4, 5, 0]
+ *     
+ */
+
+static VALUE
+rb_ary_drop(VALUE ary, VALUE n)
+{
+    VALUE result;
+    long pos = NUM2LONG(n);
+    if (pos < 0) {
+	rb_raise(rb_eArgError, "attempt to drop negative size");
+    }
+
+    result = rb_ary_subseq(ary, pos, RARRAY_LEN(ary));
+    if (result == Qnil) result = rb_ary_new();
+    return result;
+}
+
+/*
+ *  call-seq:
+ *     ary.drop_while {|arr| block }   => array
+ *  
+ *  Drops elements up to, but not including, the first element for
+ *  which the block returns nil or false and returns an array
+ *  containing the remaining elements.
+ *     
+ *     a = [1, 2, 3, 4, 5, 0]
+ *     a.drop_while {|i| i < 3 }   # => [3, 4, 5, 0]
+ *     
+ */
+
+static VALUE
+rb_ary_drop_while(VALUE ary)
+{
+    long i;
+
+    RETURN_ENUMERATOR(ary, 0, 0);
+    for (i = 0; i < RARRAY_LEN(ary); i++) {
+	if (!RTEST(rb_yield(RARRAY_PTR(ary)[i]))) break;
+    }
+    return rb_ary_drop(ary, LONG2FIX(i));
+}
+
 
 
 /* Arrays are ordered, integer-indexed collections of any object. 
@@ -3238,6 +3401,8 @@ rb_ary_product(int argc, VALUE *argv, VALUE ary)
 void
 Init_Array(void)
 {
+#undef rb_intern
+
     rb_cArray  = rb_define_class("Array", rb_cObject);
     rb_include_module(rb_cArray, rb_mEnumerable);
 
@@ -3276,6 +3441,7 @@ Init_Array(void)
     rb_define_method(rb_cArray, "length", rb_ary_length, 0);
     rb_define_alias(rb_cArray,  "size", "length");
     rb_define_method(rb_cArray, "empty?", rb_ary_empty_p, 0);
+    rb_define_method(rb_cArray, "find_index", rb_ary_index, -1);
     rb_define_method(rb_cArray, "index", rb_ary_index, -1);
     rb_define_method(rb_cArray, "rindex", rb_ary_rindex, -1);
     rb_define_method(rb_cArray, "join", rb_ary_join_m, -1);
@@ -3321,14 +3487,19 @@ Init_Array(void)
     rb_define_method(rb_cArray, "compact!", rb_ary_compact_bang, 0);
     rb_define_method(rb_cArray, "flatten", rb_ary_flatten, -1);
     rb_define_method(rb_cArray, "flatten!", rb_ary_flatten_bang, -1);
-    rb_define_method(rb_cArray, "nitems", rb_ary_nitems, 0);
+    rb_define_method(rb_cArray, "count", rb_ary_count, -1);
     rb_define_method(rb_cArray, "shuffle!", rb_ary_shuffle_bang, 0);
     rb_define_method(rb_cArray, "shuffle", rb_ary_shuffle, 0);
     rb_define_method(rb_cArray, "choice", rb_ary_choice, 0);
-    rb_define_method(rb_cArray, "cycle", rb_ary_cycle, 0);
+    rb_define_method(rb_cArray, "cycle", rb_ary_cycle, -1);
     rb_define_method(rb_cArray, "permutation", rb_ary_permutation, -1);
     rb_define_method(rb_cArray, "combination", rb_ary_combination, 1);
     rb_define_method(rb_cArray, "product", rb_ary_product, -1);
 
+    rb_define_method(rb_cArray, "take", rb_ary_take, 1);
+    rb_define_method(rb_cArray, "take_while", rb_ary_take_while, 0);
+    rb_define_method(rb_cArray, "drop", rb_ary_drop, 1);
+    rb_define_method(rb_cArray, "drop_while", rb_ary_drop_while, 0);
+
     id_cmp = rb_intern("<=>");
 }
