@@ -65,8 +65,14 @@
 #define u_short unsigned short
 #define u_int unsigned int
 
-#undef __P
+#if !defined(HAVE_STDARG_PROTOTYPES)
 #if defined(__STDC__)
+#define HAVE_STDARG_PROTOTYPES 1
+#endif
+#endif
+
+#undef __P
+#if defined(HAVE_STDARG_PROTOTYPES)
 # include <stdarg.h>
 # if !defined(__P)
 #  define __P(x) x
@@ -105,6 +111,13 @@
 #endif /* People who don't like const sys_error */
 
 #include <stddef.h>
+#if defined(__hpux) && !defined(__GNUC__) || defined(__DECC)
+#include <string.h>
+#endif
+
+#if !defined(__CYGWIN32__) && defined(__hpux) && !defined(__GNUC__)
+#include <stdlib.h>
+#endif
 
 #ifndef NULL
 #define	NULL	0
@@ -151,12 +164,15 @@ struct __sbuf {
  */
 typedef	struct __sFILE {
 	unsigned char *_p;	/* current position in (some) buffer */
+#if 0
 	int	_r;		/* read space left for getc() */
+#endif
 	int	_w;		/* write space left for putc() */
 	short	_flags;		/* flags, below; this FILE is free if 0 */
 	short	_file;		/* fileno, if Unix descriptor, else -1 */
 	struct	__sbuf _bf;	/* the buffer (at least 1 byte, if !NULL) */
 	int	_lbfsize;	/* 0 or -_bf._size, for inline putc */
+	int	(*vwrite)(/* struct __sFILE*, struct __suio * */);
 } FILE;
 
 
@@ -185,6 +201,9 @@ typedef	struct __sFILE {
 #define	__sclearerr(p)	((void)((p)->_flags &= ~(__SERR|__SEOF)))
 #define	__sfileno(p)	((p)->_file)
 
+#undef feof
+#undef ferror
+#undef clearerr
 #define	feof(p)		__sfeof(p)
 #define	ferror(p)	__sferror(p)
 #define	clearerr(p)	__sclearerr(p)
@@ -194,15 +213,11 @@ typedef	struct __sFILE {
 #endif
 
 
-#if defined(__hpux) && !defined(__GNUC__) || defined(__DECC)
-#include <string.h>
-#endif
-
 /*
  * I/O descriptors for __sfvwrite().
  */
 struct __siov {
-	void	*iov_base;
+	const void *iov_base;
 	size_t	iov_len;
 };
 struct __suio {
@@ -211,13 +226,14 @@ struct __suio {
 	int	uio_resid;
 };
 
+#if !defined(HAVE_VSNPRINTF) || !defined(HAVE_SNPRINTF)
 /*
  * Write some memory regions.  Return zero on success, EOF on error.
  *
  * This routine is large and unsightly, but most of the ugliness due
  * to the three different kinds of output buffering is handled here.
  */
-static BSD__sfvwrite(fp, uio)
+static int BSD__sfvwrite(fp, uio)
 	register FILE *fp;
 	register struct __suio *uio;
 {
@@ -284,11 +300,8 @@ static BSD__sfvwrite(fp, uio)
 		*/
 	}
 	return (0);
-
-err:
-	fp->_flags |= __SERR;
-	return (EOF);
 }
+#endif
 
 /*
  * Actual printf innards.
@@ -296,18 +309,12 @@ err:
  * This code is large and complicated...
  */
 
-#if !defined(__CYGWIN32__) && defined(__hpux) && !defined(__GNUC__)
-#include <stdlib.h>
-#endif
-
 /*
  * Flush out all the vectors defined by the given uio,
  * then reset it so that it can be reused.
  */
 static int
-BSD__sprint(fp, uio)
-	FILE *fp;
-	register struct __suio *uio;
+BSD__sprint(FILE *fp, register struct __suio *uio)
 {
 	register int err;
 
@@ -315,7 +322,7 @@ BSD__sprint(fp, uio)
 		uio->uio_iovcnt = 0;
 		return (0);
 	}
-	err = BSD__sfvwrite(fp, uio);
+	err = (*fp->vwrite)(fp, uio);
 	uio->uio_resid = 0;
 	uio->uio_iovcnt = 0;
 	return (err);
@@ -328,10 +335,7 @@ BSD__sprint(fp, uio)
  * worries about ungetc buffers and so forth.
  */
 static int
-BSD__sbprintf(fp, fmt, ap)
-	register FILE *fp;
-	const char *fmt;
-	va_list ap;
+BSD__sbprintf(register FILE *fp, const char *fmt, va_list ap)
 {
 /* We don't support files. */
 	return 0;
@@ -345,6 +349,72 @@ BSD__sbprintf(fp, fmt, ap)
 #define is_digit(c)	((unsigned)to_digit(c) <= 9)
 #define	to_char(n)	((n) + '0')
 
+#ifdef _HAVE_SANE_QUAD_
+/*
+ * Convert an unsigned long long to ASCII for printf purposes, returning
+ * a pointer to the first character of the string representation.
+ * Octal numbers can be forced to have a leading zero; hex numbers
+ * use the given digits.
+ */
+static char *
+BSD__uqtoa(register u_quad_t val, char *endp, int base, int octzero, char *xdigs)
+{
+	register char *cp = endp;
+	register long sval;
+
+	/*
+	 * Handle the three cases separately, in the hope of getting
+	 * better/faster code.
+	 */
+	switch (base) {
+	case 10:
+		if (val < 10) {	/* many numbers are 1 digit */
+			*--cp = to_char(val);
+			return (cp);
+		}
+		/*
+		 * On many machines, unsigned arithmetic is harder than
+		 * signed arithmetic, so we do at most one unsigned mod and
+		 * divide; this is sufficient to reduce the range of
+		 * the incoming value to where signed arithmetic works.
+		 */
+		if (val > LLONG_MAX) {
+			*--cp = to_char(val % 10);
+			sval = val / 10;
+		} else
+			sval = val;
+		do {
+			*--cp = to_char(sval % 10);
+			sval /= 10;
+		} while (sval != 0);
+		break;
+
+	case 8:
+		do {
+			*--cp = to_char(val & 7);
+			val >>= 3;
+		} while (val);
+		if (octzero && *cp != '0')
+			*--cp = '0';
+		break;
+
+	case 16:
+		do {
+			*--cp = xdigs[val & 15];
+			val >>= 4;
+		} while (val);
+		break;
+
+	default:			/* oops */
+		/* 
+		abort();
+		*/
+		break;	/* fjc 7-31-97.  Don't reference abort() here */
+	}
+	return (cp);
+}
+#endif /* _HAVE_SANE_QUAD_ */
+
 /*
  * Convert an unsigned long to ASCII for printf purposes, returning
  * a pointer to the first character of the string representation.
@@ -352,11 +422,7 @@ BSD__sbprintf(fp, fmt, ap)
  * use the given digits.
  */
 static char *
-BSD__ultoa(val, endp, base, octzero, xdigs)
-	register u_long val;
-	char *endp;
-	int base, octzero;
-	char *xdigs;
+BSD__ultoa(register u_long val, char *endp, int base, int octzero, const char *xdigs)
 {
 	register char *cp = endp;
 	register long sval;
@@ -455,15 +521,12 @@ static int exponent __P((char *, int, int));
 #define	ZEROPAD		0x080		/* zero (as opposed to blank) pad */
 #define FPT		0x100		/* Floating point number */
 static int
-BSD_vfprintf(fp, fmt0, ap)
-	FILE *fp;
-	const char *fmt0;
-	va_list ap;
+BSD_vfprintf(FILE *fp, const char *fmt0, va_list ap)
 {
-	register char *fmt;	/* format string */
+	register const char *fmt; /* format string */
 	register int ch;	/* character from fmt */
 	register int n;		/* handy integer (short term usage) */
-	register char *cp;	/* handy char pointer (short term usage) */
+	register const char *cp;/* handy char pointer (short term usage) */
 	register struct __siov *iovp;/* for PRINT macro */
 	register int flags;	/* flags as above */
 	int ret;		/* return value accumulator */
@@ -487,12 +550,13 @@ BSD_vfprintf(fp, fmt0, ap)
 	int fieldsz;		/* field size expanded by sign, etc */
 	int realsz;		/* field size expanded by dprec */
 	int size;		/* size of converted field or string */
-	char *xdigs;		/* digits for [xX] conversion */
+	const char *xdigs = 0;	/* digits for [xX] conversion */
 #define NIOV 8
 	struct __suio uio;	/* output information: summary */
 	struct __siov iov[NIOV];/* ... and individual io vectors */
 	char buf[BUF];		/* space for %c, %[diouxX], %[eEfgG] */
 	char ox[2];		/* space for 0x hex-prefix */
+	char *const ebuf = buf + sizeof(buf);
 
 	/*
 	 * Choose PADSIZE to trade efficiency vs. size.  If larger printf
@@ -553,7 +617,7 @@ BSD_vfprintf(fp, fmt0, ap)
 	    fp->_file >= 0)
 		return (BSD__sbprintf(fp, fmt0, ap));
 
-	fmt = (char *)fmt0;
+	fmt = fmt0;
 	uio.uio_iov = iovp = iov;
 	uio.uio_resid = 0;
 	uio.uio_iovcnt = 0;
@@ -657,7 +721,8 @@ reswitch:	switch (ch) {
 			goto rflag;
 #endif /* _HAVE_SANE_QUAD_ */
 		case 'c':
-			*(cp = buf) = va_arg(ap, int);
+			cp = buf;
+			*buf = (char)va_arg(ap, int);
 			size = 1;
 			sign = '\0';
 			break;
@@ -673,10 +738,9 @@ reswitch:	switch (ch) {
 					uqval = -uqval;
 					sign = '-';
 				}
-			} else {
-#else /* _HAVE_SANE_QUAD_ */
-			{
+			} else
 #endif /* _HAVE_SANE_QUAD_ */
+			{
 				ulval = SARG();
 				if ((long)ulval < 0) {
 					ulval = -ulval;
@@ -781,6 +845,7 @@ fp_begin:		_double = va_arg(ap, double);
 			 * defined manner.''
 			 *	-- ANSI X3J11
 			 */
+		        prec = sizeof(void*)*CHAR_BIT/4;
 #ifdef _HAVE_LLP64_
 			uqval = (u_long)va_arg(ap, void *);
 			flags = (flags) | QUADINT | HEXPREFIX;
@@ -805,7 +870,7 @@ fp_begin:		_double = va_arg(ap, double);
 				 * NUL in the first `prec' characters, and
 				 * strlen() will go further.
 				 */
-				char *p = (char *)memchr(cp, 0, prec);
+				const char *p = (char *)memchr(cp, 0, prec);
 
 				if (p != NULL) {
 					size = p - cp;
@@ -845,10 +910,11 @@ hex:
 			/* leading 0x/X only if non-zero */
 			if (flags & ALT &&
 #ifdef _HAVE_SANE_QUAD_
-			    (flags & QUADINT ? uqval != 0 : ulval != 0))
+			    (flags & QUADINT ? uqval != 0 : ulval != 0)
 #else /* _HAVE_SANE_QUAD_ */
-			    ulval != 0)
+			    ulval != 0
 #endif /* _HAVE_SANE_QUAD_ */
+			    )
 				flags |= HEXPREFIX;
 
 			/* unsigned conversions */
@@ -866,28 +932,27 @@ number:			if ((dprec = prec) >= 0)
 			 * explicit precision of zero is no characters.''
 			 *	-- ANSI X3J11
 			 */
-			cp = buf + BUF;
 #ifdef _HAVE_SANE_QUAD_
 			if (flags & QUADINT) {
 				if (uqval != 0 || prec != 0)
-					cp = __uqtoa(uqval, cp, base,
+					cp = BSD__uqtoa(uqval, ebuf, base,
 					    flags & ALT, xdigs);
-			} else {
+			} else
 #else /* _HAVE_SANE_QUAD_ */
-			{
 #endif /* _HAVE_SANE_QUAD_ */
+			{
 				if (ulval != 0 || prec != 0)
-					cp = BSD__ultoa(ulval, cp, base,
+					cp = BSD__ultoa(ulval, ebuf, base,
 					    flags & ALT, xdigs);
 			}
-			size = buf + BUF - cp;
+			size = ebuf - cp;
 			break;
 		default:	/* "%?" prints ?, unless ? is NUL */
 			if (ch == '\0')
 				goto done;
 			/* pretend it was %c with argument ch */
 			cp = buf;
-			*cp = ch;
+			*buf = ch;
 			size = 1;
 			sign = '\0';
 			break;
@@ -1073,6 +1138,7 @@ exponent(p0, exp, fmtch)
 }
 #endif /* FLOATING_POINT */
 
+#ifndef HAVE_VSNPRINTF
 int
 vsnprintf(str, n, fmt, ap)
 	char *str;
@@ -1088,23 +1154,20 @@ vsnprintf(str, n, fmt, ap)
 	f._flags = __SWR | __SSTR;
 	f._bf._base = f._p = (unsigned char *)str;
 	f._bf._size = f._w = n - 1;
+	f.vwrite = BSD__sfvwrite;
 	ret = BSD_vfprintf(&f, fmt, ap);
 	*f._p = 0;
 	return (ret);
 }
+#endif
 
 #if defined(LIBC_SCCS) && !defined(lint)
 static char sccsid[] = "@(#)snprintf.c	8.1 (Berkeley) 6/4/93";
 #endif /* LIBC_SCCS and not lint */
 
-#if defined(__STDC__)
-# include <stdarg.h>
-#else
-# include <varargs.h>
-#endif
-
+#ifndef HAVE_SNPRINTF
 int
-#if defined(__STDC__)
+#if defined(HAVE_STDARG_PROTOTYPES)
 snprintf(char *str, size_t n, char const *fmt, ...)
 #else
 snprintf(str, n, fmt, va_alist)
@@ -1120,7 +1183,7 @@ va_dcl
 	if ((int)n < 1)
 		return (EOF);
 
-#if defined(__STDC__)
+#if defined(HAVE_STDARG_PROTOTYPES)
 	va_start(ap, fmt);
 #else
 	va_start(ap);
@@ -1128,8 +1191,10 @@ va_dcl
 	f._flags = __SWR | __SSTR;
 	f._bf._base = f._p = (unsigned char *)str;
 	f._bf._size = f._w = n - 1;
+	f.vwrite = BSD__sfvwrite;
 	ret = BSD_vfprintf(&f, fmt, ap);
 	*f._p = 0;
 	va_end(ap);
 	return (ret);
 }
+#endif
