@@ -2,7 +2,7 @@
 
   thread.c -
 
-  $Author: matz $
+  $Author: akr $
 
   Copyright (C) 2004-2007 Koichi Sasada
 
@@ -57,11 +57,14 @@ VALUE rb_cBarrier;
 
 static void sleep_timeval(rb_thread_t *th, struct timeval time);
 static void sleep_wait_for_interrupt(rb_thread_t *th, double sleepsec);
-static void sleep_forever(rb_thread_t *th);
+static void sleep_forever(rb_thread_t *th, int nodeadlock);
 static double timeofday(void);
 struct timeval rb_time_interval(VALUE);
 static int rb_thread_dead(rb_thread_t *th);
 
+static void rb_mutex_unlock_all(VALUE);
+static void rb_check_deadlock(rb_vm_t *vm);
+
 void rb_signal_exec(rb_thread_t *th, int sig);
 void rb_disable_interrupt(void);
 
@@ -70,43 +73,47 @@ static VALUE eTerminateSignal = INT2FIX(1);
 static volatile int system_working = 1;
 
 inline static void
-st_delete_wrap(st_table * table, VALUE key)
+st_delete_wrap(st_table *table, st_data_t key)
 {
-    st_delete(table, (st_data_t *) & key, 0);
+    st_delete(table, &key, 0);
 }
 
 /********************************************************************************/
 
 #define THREAD_SYSTEM_DEPENDENT_IMPLEMENTATION
 
-static void set_unblock_function(rb_thread_t *th, rb_unblock_function_t *func, void *ptr,
-				 rb_unblock_function_t **oldfunc, void **oldptr);
+static void set_unblock_function(rb_thread_t *th, rb_unblock_function_t *func, void *arg,
+				 struct rb_unblock_callback *old);
+static void reset_unblock_function(rb_thread_t *th, const struct rb_unblock_callback *old);
 
 #define GVL_UNLOCK_BEGIN() do { \
   rb_thread_t *_th_stored = GET_THREAD(); \
   rb_gc_save_machine_context(_th_stored); \
-  native_mutex_unlock(&_th_stored->vm->global_interpreter_lock)
+  native_mutex_unlock(&_th_stored->vm->global_vm_lock)
 
 #define GVL_UNLOCK_END() \
-  native_mutex_lock(&_th_stored->vm->global_interpreter_lock); \
+  native_mutex_lock(&_th_stored->vm->global_vm_lock); \
   rb_thread_set_current(_th_stored); \
 } while(0)
 
-#define BLOCKING_REGION(exec, ubf, ubfarg) do { \
-    rb_thread_t *__th = GET_THREAD(); \
-    int __prev_status = __th->status; \
-    rb_unblock_function_t *__oldubf; \
-    void *__oldubfarg; \
-    set_unblock_function(__th, ubf, ubfarg, &__oldubf, &__oldubfarg); \
-    __th->status = THREAD_STOPPED; \
-    thread_debug("enter blocking region (%p)\n", __th); \
+#define BLOCKING_REGION_CORE(exec) do { \
     GVL_UNLOCK_BEGIN(); {\
 	    exec; \
     } \
     GVL_UNLOCK_END(); \
+} while(0);
+
+#define BLOCKING_REGION(exec, ubf, ubfarg) do { \
+    rb_thread_t *__th = GET_THREAD(); \
+    int __prev_status = __th->status; \
+    struct rb_unblock_callback __oldubf; \
+    set_unblock_function(__th, ubf, ubfarg, &__oldubf); \
+    __th->status = THREAD_STOPPED; \
+    thread_debug("enter blocking region (%p)\n", __th); \
+    BLOCKING_REGION_CORE(exec); \
     thread_debug("leave blocking region (%p)\n", __th); \
     remove_signal_thread_list(__th); \
-    set_unblock_function(__th, __oldubf, __oldubfarg, 0, 0); \
+    reset_unblock_function(__th, &__oldubf); \
     if (__th->status == THREAD_STOPPED) { \
 	__th->status = __prev_status; \
     } \
@@ -195,7 +202,7 @@ rb_thread_debug(const char *fmt, ...)
 
 static void
 set_unblock_function(rb_thread_t *th, rb_unblock_function_t *func, void *arg,
-		     rb_unblock_function_t **oldfunc, void **oldarg)
+		     struct rb_unblock_callback *old)
 {
   check_ints:
     RUBY_VM_CHECK_INTS(); /* check signal or so */
@@ -205,21 +212,28 @@ set_unblock_function(rb_thread_t *th, rb_unblock_function_t *func, void *arg,
 	goto check_ints;
     }
     else {
-	if (oldfunc) *oldfunc = th->unblock_function;
-	if (oldarg) *oldarg = th->unblock_function_arg;
-	th->unblock_function = func;
-	th->unblock_function_arg = arg;
+	if (old) *old = th->unblock;
+	th->unblock.func = func;
+	th->unblock.arg = arg;
     }
     native_mutex_unlock(&th->interrupt_lock);
 }
 
+static void
+reset_unblock_function(rb_thread_t *th, const struct rb_unblock_callback *old)
+{
+    native_mutex_lock(&th->interrupt_lock);
+    th->unblock = *old;
+    native_mutex_unlock(&th->interrupt_lock);
+}
+
 static void
 rb_thread_interrupt(rb_thread_t *th)
 {
     native_mutex_lock(&th->interrupt_lock);
     RUBY_VM_SET_INTERRUPT(th);
-    if (th->unblock_function) {
-	(th->unblock_function)(th->unblock_function_arg);
+    if (th->unblock.func) {
+	(th->unblock.func)(th->unblock.arg);
     }
     else {
 	/* none */
@@ -256,6 +270,11 @@ rb_thread_terminate_all(void)
 	rb_bug("rb_thread_terminate_all: called by child thread (%p, %p)", vm->main_thread, th);
     }
 
+    /* unlock all locking mutexes */
+    if (th->keeping_mutexes) {
+	rb_mutex_unlock_all(th->keeping_mutexes);
+    }
+
     thread_debug("rb_thread_terminate_all (main thread: %p)\n", th);
     st_foreach(vm->living_threads, terminate_i, (st_data_t)th);
 
@@ -273,7 +292,7 @@ rb_thread_terminate_all(void)
 }
 
 static void
-thread_cleanup_func(void *th_ptr)
+thread_cleanup_func_before_exec(void *th_ptr)
 {
     rb_thread_t *th = th_ptr;
     th->status = THREAD_KILLED;
@@ -281,6 +300,13 @@ thread_cleanup_func(void *th_ptr)
 #ifdef __ia64
     th->machine_register_stack_start = th->machine_register_stack_end = 0;
 #endif
+}
+
+static void
+thread_cleanup_func(void *th_ptr)
+{
+    rb_thread_t *th = th_ptr;
+    thread_cleanup_func_before_exec(th_ptr);
     native_thread_destroy(th);
 }
 
@@ -288,6 +314,12 @@ extern void ruby_error_print(void);
 static VALUE rb_thread_raise(int, VALUE *, rb_thread_t *);
 void rb_thread_recycle_stack_release(VALUE *);
 
+void
+ruby_thread_init_stack(rb_thread_t *th)
+{
+    native_thread_init_stack(th);
+}
+
 static int
 thread_start_func_2(rb_thread_t *th, VALUE *stack_start, VALUE *register_stack_start)
 {
@@ -304,7 +336,7 @@ thread_start_func_2(rb_thread_t *th, VALUE *stack_start, VALUE *register_stack_s
 #endif
     thread_debug("thread start: %p\n", th);
 
-    native_mutex_lock(&th->vm->global_interpreter_lock);
+    native_mutex_lock(&th->vm->global_vm_lock);
     {
 	thread_debug("thread start (get lock): %p\n", th);
 	rb_thread_set_current(th);
@@ -347,6 +379,19 @@ thread_start_func_2(rb_thread_t *th, VALUE *stack_start, VALUE *register_stack_s
 	}
 	TH_POP_TAG();
 
+	/* locking_mutex must be Qfalse */
+	if (th->locking_mutex != Qfalse) {
+	    rb_bug("thread_start_func_2: locking_mutex must not be set (%p:%"PRIxVALUE")",
+		   th, th->locking_mutex);
+	}
+
+	/* unlock all locking mutexes */
+	if (th->keeping_mutexes) {
+	    rb_mutex_unlock_all(th->keeping_mutexes);
+	    th->keeping_mutexes = Qfalse;
+	}
+
+	/* delete self from living_threads */
 	st_delete_wrap(th->vm->living_threads, th->self);
 
 	/* wake up joinning threads */
@@ -356,7 +401,7 @@ thread_start_func_2(rb_thread_t *th, VALUE *stack_start, VALUE *register_stack_s
 	    rb_thread_interrupt(join_th);
 	    join_th = join_th->join_list_next;
 	}
-	st_delete_wrap(th->vm->living_threads, th->self);
+	if (th != main_th) rb_check_deadlock(th->vm);
 
 	if (!th->root_fiber) {
 	    rb_thread_recycle_stack_release(th->stack);
@@ -364,7 +409,7 @@ thread_start_func_2(rb_thread_t *th, VALUE *stack_start, VALUE *register_stack_s
 	}
     }
     thread_cleanup_func(th);
-    native_mutex_unlock(&th->vm->global_interpreter_lock);
+    native_mutex_unlock(&th->vm->global_vm_lock);
 
     return 0;
 }
@@ -461,34 +506,80 @@ rb_thread_create(VALUE (*fn)(ANYARGS), void *arg)
 /* +infty, for this purpose */
 #define DELAY_INFTY 1E30
 
-static VALUE
-thread_join(rb_thread_t *target_th, double delay)
-{
-    rb_thread_t *th = GET_THREAD();
-    double now, limit = timeofday() + delay;
+struct join_arg {
+    rb_thread_t *target, *waiting;
+    double limit;
+    int forever;
+};
 
-    thread_debug("thread_join (thid: %p)\n", (void *)target_th->thread_id);
+static VALUE
+remove_from_join_list(VALUE arg)
+{
+    struct join_arg *p = (struct join_arg *)arg;
+    rb_thread_t *target_th = p->target, *th = p->waiting;
 
     if (target_th->status != THREAD_KILLED) {
-	th->join_list_next = target_th->join_list_head;
-	target_th->join_list_head = th;
+	rb_thread_t **pth = &target_th->join_list_head;
+
+	while (*pth) {
+	    if (*pth == th) {
+		*pth = th->join_list_next;
+		break;
+	    }
+	    pth = &(*pth)->join_list_next;
+	}
     }
+
+    return Qnil;
+}
+
+static VALUE
+thread_join_sleep(VALUE arg)
+{
+    struct join_arg *p = (struct join_arg *)arg;
+    rb_thread_t *target_th = p->target, *th = p->waiting;
+    double now, limit = p->limit;
+
     while (target_th->status != THREAD_KILLED) {
-	if (delay == DELAY_INFTY) {
-	    sleep_forever(th);
+	if (p->forever) {
+	    sleep_forever(th, 1);
 	}
 	else {
 	    now = timeofday();
 	    if (now > limit) {
 		thread_debug("thread_join: timeout (thid: %p)\n",
 			     (void *)target_th->thread_id);
-		return Qnil;
+		return Qfalse;
 	    }
 	    sleep_wait_for_interrupt(th, limit - now);
 	}
 	thread_debug("thread_join: interrupted (thid: %p)\n",
 		     (void *)target_th->thread_id);
     }
+    return Qtrue;
+}
+
+static VALUE
+thread_join(rb_thread_t *target_th, double delay)
+{
+    rb_thread_t *th = GET_THREAD();
+    struct join_arg arg;
+
+    arg.target = target_th;
+    arg.waiting = th;
+    arg.limit = timeofday() + delay;
+    arg.forever = delay == DELAY_INFTY;
+
+    thread_debug("thread_join (thid: %p)\n", (void *)target_th->thread_id);
+
+    if (target_th->status != THREAD_KILLED) {
+	th->join_list_next = target_th->join_list_head;
+	target_th->join_list_head = th;
+	if (!rb_ensure(thread_join_sleep, (VALUE)&arg,
+		       remove_from_join_list, (VALUE)&arg)) {
+	    return Qnil;
+	}
+    }
 
     thread_debug("thread_join: success (thid: %p)\n",
 		 (void *)target_th->thread_id);
@@ -607,30 +698,46 @@ double2timeval(double d)
 }
 
 static void
-sleep_forever(rb_thread_t *th)
+sleep_forever(rb_thread_t *th, int deadlockable)
 {
-    native_sleep(th, 0);
+    native_sleep(th, 0, deadlockable);
 }
 
 static void
 sleep_timeval(rb_thread_t *th, struct timeval tv)
 {
-    native_sleep(th, &tv);
+    native_sleep(th, &tv, 0);
 }
 
 void
 rb_thread_sleep_forever()
 {
     thread_debug("rb_thread_sleep_forever\n");
-    sleep_forever(GET_THREAD());
+    sleep_forever(GET_THREAD(), 0);
+}
+
+static void
+rb_thread_sleep_deadly()
+{
+    thread_debug("rb_thread_sleep_deadly\n");
+    sleep_forever(GET_THREAD(), 1);
 }
 
 static double
 timeofday(void)
 {
-    struct timeval tv;
-    gettimeofday(&tv, NULL);
-    return (double)tv.tv_sec + (double)tv.tv_usec * 1e-6;
+#if defined(HAVE_CLOCK_GETTIME) && defined(CLOCK_MONOTONIC)
+    struct timespec tp;
+
+    if (clock_gettime(CLOCK_MONOTONIC, &tp) == 0) {
+        return (double)tp.tv_sec + (double)tp.tv_nsec * 1e-9;
+    } else
+#endif
+    {
+        struct timeval tv;
+        gettimeofday(&tv, NULL);
+        return (double)tv.tv_sec + (double)tv.tv_usec * 1e-6;
+    }
 }
 
 static void
@@ -683,11 +790,11 @@ rb_thread_schedule(void)
 	thread_debug("rb_thread_schedule/switch start\n");
 
 	rb_gc_save_machine_context(th);
-	native_mutex_unlock(&th->vm->global_interpreter_lock);
+	native_mutex_unlock(&th->vm->global_vm_lock);
 	{
 	    native_thread_yield();
 	}
-	native_mutex_lock(&th->vm->global_interpreter_lock);
+	native_mutex_lock(&th->vm->global_vm_lock);
 
 	rb_thread_set_current(th);
 	thread_debug("rb_thread_schedule/switch done\n");
@@ -752,6 +859,7 @@ thread_s_pass(VALUE klass)
 void
 rb_thread_execute_interrupts(rb_thread_t *th)
 {
+    if (th->raised_flag) return;
     while (th->interrupt_flag) {
 	int status = th->status;
 	th->status = THREAD_RUNNABLE;
@@ -1058,7 +1166,7 @@ rb_thread_stop(void)
 	rb_raise(rb_eThreadError,
 		 "stopping only thread\n\tnote: use sleep to stop forever");
     }
-    rb_thread_sleep_forever();
+    rb_thread_sleep_deadly();
     return Qnil;
 }
 
@@ -1072,6 +1180,7 @@ thread_list_i(st_data_t key, st_data_t val, void *data)
     switch (th->status) {
       case THREAD_RUNNABLE:
       case THREAD_STOPPED:
+      case THREAD_STOPPED_FOREVER:
       case THREAD_TO_KILL:
 	rb_ary_push(ary, th->self);
       default:
@@ -1266,6 +1375,7 @@ thread_status_name(enum rb_thread_status status)
       case THREAD_RUNNABLE:
 	return "run";
       case THREAD_STOPPED:
+      case THREAD_STOPPED_FOREVER:
 	return "sleep";
       case THREAD_TO_KILL:
 	return "aborting";
@@ -1365,7 +1475,7 @@ rb_thread_stop_p(VALUE thread)
 
     if (rb_thread_dead(th))
 	return Qtrue;
-    if (th->status == THREAD_STOPPED)
+    if (th->status == THREAD_STOPPED || th->status == THREAD_STOPPED_FOREVER)
 	return Qtrue;
     return Qfalse;
 }
@@ -1401,7 +1511,7 @@ rb_thread_safe_level(VALUE thread)
 static VALUE
 rb_thread_inspect(VALUE thread)
 {
-    char *cname = rb_obj_classname(thread);
+    const char *cname = rb_obj_classname(thread);
     rb_thread_t *th;
     const char *status;
     VALUE str;
@@ -1476,7 +1586,7 @@ rb_thread_local_aset(VALUE thread, ID id, VALUE val)
 	th->local_storage = st_init_numtable();
     }
     if (NIL_P(val)) {
-	st_delete(th->local_storage, (st_data_t *) & id, 0);
+	st_delete_wrap(th->local_storage, id);
 	return Qnil;
     }
     st_insert(th->local_storage, id, val);
@@ -1511,15 +1621,17 @@ rb_thread_aset(VALUE self, ID id, VALUE val)
  */
 
 static VALUE
-rb_thread_key_p(VALUE self, ID id)
+rb_thread_key_p(VALUE self, VALUE key)
 {
     rb_thread_t *th;
+    ID id = rb_to_id(key);
+
     GetThreadPtr(self, th);
 
     if (!th->local_storage) {
 	return Qfalse;
     }
-    if (st_lookup(th->local_storage, rb_to_id(id), 0)) {
+    if (st_lookup(th->local_storage, id, 0)) {
 	return Qtrue;
     }
     return Qfalse;
@@ -1532,12 +1644,18 @@ thread_keys_i(ID key, VALUE value, VALUE ary)
     return ST_CONTINUE;
 }
 
+static int
+vm_living_thread_num(rb_vm_t *vm)
+{
+    return vm->living_threads->num_entries;
+}
+
 int
 rb_thread_alone()
 {
     int num = 1;
     if (GET_THREAD()->vm->living_threads) {
-	num = GET_THREAD()->vm->living_threads->num_entries;
+	num = vm_living_thread_num(GET_THREAD()->vm);
 	thread_debug("rb_thread_alone: %d\n", num);
     }
     return num == 1;
@@ -1640,7 +1758,7 @@ rb_fd_init(volatile rb_fdset_t *fds)
 void
 rb_fd_term(rb_fdset_t *fds)
 {
-    if (fds->fdset) free(fds->fdset);
+    if (fds->fdset) xfree(fds->fdset);
     fds->maxfd = 0;
     fds->fdset = 0;
 }
@@ -1702,6 +1820,25 @@ rb_fd_copy(rb_fdset_t *dst, const fd_set *src, int max)
     memcpy(dst->fdset, src, size);
 }
 
+int
+rb_fd_select(int n, rb_fdset_t *readfds, rb_fdset_t *writefds, rb_fdset_t *exceptfds, struct timeval *timeout)
+{
+    fd_set *r = NULL, *w = NULL, *e = NULL;
+    if (readfds) {
+        rb_fd_resize(n - 1, readfds);
+        r = rb_fd_ptr(readfds);
+    }
+    if (writefds) {
+        rb_fd_resize(n - 1, writefds);
+        w = rb_fd_ptr(writefds);
+    }
+    if (exceptfds) {
+        rb_fd_resize(n - 1, exceptfds);
+        e = rb_fd_ptr(exceptfds);
+    }
+    return select(n, r, w, e, timeout);
+}
+
 #undef FD_ZERO
 #undef FD_SET
 #undef FD_CLR
@@ -1746,7 +1883,7 @@ do_select(int n, fd_set *read, fd_set *write, fd_set *except,
     fd_set orig_read, orig_write, orig_except;
 
 #ifndef linux
-    double limit;
+    double limit = 0;
     struct timeval wait_rest;
 
     if (timeout) {
@@ -1797,11 +1934,11 @@ do_select(int n, fd_set *read, fd_set *write, fd_set *except,
     errno = lerrno;
 
     if (result < 0) {
-	if (errno == EINTR
+	switch (errno) {
+	  case EINTR:
 #ifdef ERESTART
-	    || errno == ERESTART
+	  case ERESTART:
 #endif
-	    ) {
 	    if (read) *read = orig_read;
 	    if (write) *write = orig_write;
 	    if (except) *except = orig_except;
@@ -1816,6 +1953,8 @@ do_select(int n, fd_set *read, fd_set *write, fd_set *except,
 	    }
 #endif
 	    goto retry;
+	  default:
+	    break;
 	}
     }
     return result;
@@ -1897,6 +2036,7 @@ void
 rb_gc_save_machine_context(rb_thread_t *th)
 {
     SET_MACHINE_STACK_END(&th->machine_stack_end);
+    FLUSH_REGISTER_WINDOWS;
 #ifdef __ia64
     th->machine_register_stack_end = rb_ia64_bsp();
 #endif
@@ -1920,8 +2060,8 @@ timer_thread_function(void)
     /* check signal */
     if (vm->buffered_signal_size && vm->main_thread->exec_signal == 0) {
 	vm->main_thread->exec_signal = rb_get_next_signal(vm);
-	thread_debug("buffered_signal_size: %d, sig: %d\n",
-		     vm->buffered_signal_size, vm->main_thread->exec_signal);
+	thread_debug("buffered_signal_size: %ld, sig: %d\n",
+		     (long)vm->buffered_signal_size, vm->main_thread->exec_signal);
 	rb_thread_interrupt(vm->main_thread);
     }
 
@@ -1943,6 +2083,7 @@ rb_thread_stop_timer_thread(void)
     if (timer_thread_id) {
 	system_working = 0;
 	native_thread_join(timer_thread_id);
+	timer_thread_id = 0;
     }
 }
 
@@ -1982,6 +2123,35 @@ rb_thread_atfork(void)
     st_foreach(vm->living_threads, terminate_atfork_i, (st_data_t)th);
     st_clear(vm->living_threads);
     st_insert(vm->living_threads, thval, (st_data_t) th->thread_id);
+    vm->sleeper = 0;
+    rb_reset_random_seed();
+}
+
+static int
+terminate_atfork_before_exec_i(st_data_t key, st_data_t val, rb_thread_t *current_th)
+{
+    VALUE thval = key;
+    rb_thread_t *th;
+    GetThreadPtr(thval, th);
+
+    if (th != current_th) {
+	thread_cleanup_func_before_exec(th);
+    }
+    return ST_CONTINUE;
+}
+
+void
+rb_thread_atfork_before_exec(void)
+{
+    rb_thread_t *th = GET_THREAD();
+    rb_vm_t *vm = th->vm;
+    VALUE thval = th->self;
+    vm->main_thread = th;
+
+    st_foreach(vm->living_threads, terminate_atfork_before_exec_i, (st_data_t)th);
+    st_clear(vm->living_threads);
+    st_insert(vm->living_threads, thval, (st_data_t) th->thread_id);
+    vm->sleeper = 0;
 }
 
 struct thgroup {
@@ -2008,7 +2178,7 @@ thgroup_s_alloc(VALUE klass)
     VALUE group;
     struct thgroup *data;
 
-    group = Data_Make_Struct(klass, struct thgroup, 0, free, data);
+    group = Data_Make_Struct(klass, struct thgroup, 0, -1, data);
     data->enclosed = 0;
     data->group = group;
 
@@ -2198,12 +2368,15 @@ typedef struct mutex_struct {
     rb_thread_lock_t lock;
     rb_thread_cond_t cond;
     rb_thread_t volatile *th;
-    volatile int cond_waiting;
+    volatile int cond_waiting, cond_notified;
+    VALUE next_mutex;
 } mutex_t;
 
 #define GetMutexPtr(obj, tobj) \
   Data_Get_Struct(obj, mutex_t, tobj)
 
+static const char *mutex_unlock(mutex_t *mutex);
+
 static void
 mutex_mark(void *ptr)
 {
@@ -2220,6 +2393,10 @@ mutex_free(void *ptr)
 {
     if (ptr) {
 	mutex_t *mutex = ptr;
+	if (mutex->th) {
+	    /* rb_warn("free locked mutex"); */
+	    mutex_unlock(mutex);
+	}
 	native_mutex_destroy(&mutex->lock);
 	native_cond_destroy(&mutex->cond);
     }
@@ -2270,6 +2447,17 @@ rb_mutex_locked_p(VALUE self)
     return mutex->th ? Qtrue : Qfalse;
 }
 
+static void
+mutex_locked(rb_thread_t *th, VALUE self)
+{
+    if (th->keeping_mutexes) {
+	mutex_t *mutex;
+	GetMutexPtr(self, mutex);
+	mutex->next_mutex = th->keeping_mutexes;
+    }
+    th->keeping_mutexes = self;
+}
+
 /*
  * call-seq:
  *    mutex.try_lock  => true or false
@@ -2292,40 +2480,48 @@ rb_mutex_trylock(VALUE self)
     if (mutex->th == 0) {
 	mutex->th = GET_THREAD();
 	locked = Qtrue;
+
+	mutex_locked(GET_THREAD(), self);
     }
     native_mutex_unlock(&mutex->lock);
 
     return locked;
 }
 
-static VALUE
-lock_func(rb_thread_t *th, mutex_t *mutex)
+static int
+lock_func(rb_thread_t *th, mutex_t *mutex, int last_thread)
 {
-    int locked = 0;
+    int interrupted = 0;
+#if 0 /* for debug */
+    native_thread_yield();
+#endif
 
-    while (locked == 0) {
-	native_mutex_lock(&mutex->lock);
-	{
-	    if (mutex->th == 0) {
-		mutex->th = th;
-		locked = 1;
-	    }
-	    else {
-		mutex->cond_waiting++;
-		native_cond_wait(&mutex->cond, &mutex->lock);
+    native_mutex_lock(&mutex->lock);
+    th->transition_for_lock = 0;
+    while (mutex->th || (mutex->th = th, 0)) {
+	if (last_thread) {
+	    interrupted = 2;
+	    break;
+	}
 
-		if (th->interrupt_flag) {
-		    locked = 1;
-		}
-		else if (mutex->th == 0) {
-		    mutex->th = th;
-		    locked = 1;
-		}
-	    }
+	mutex->cond_waiting++;
+	native_cond_wait(&mutex->cond, &mutex->lock);
+	mutex->cond_notified--;
+
+	if (RUBY_VM_INTERRUPTED(th)) {
+	    interrupted = 1;
+	    break;
 	}
-	native_mutex_unlock(&mutex->lock);
     }
-    return Qnil;
+    th->transition_for_lock = 1;
+    native_mutex_unlock(&mutex->lock);
+
+    if (interrupted == 2) native_thread_yield();
+#if 0 /* for debug */
+    native_thread_yield();
+#endif
+
+    return interrupted;
 }
 
 static void
@@ -2335,6 +2531,7 @@ lock_interrupt(void *ptr)
     native_mutex_lock(&mutex->lock);
     if (mutex->cond_waiting > 0) {
 	native_cond_broadcast(&mutex->cond);
+	mutex->cond_notified = mutex->cond_waiting;
 	mutex->cond_waiting = 0;
     }
     native_mutex_unlock(&mutex->lock);
@@ -2356,29 +2553,52 @@ rb_mutex_lock(VALUE self)
 	GetMutexPtr(self, mutex);
 
 	while (mutex->th != th) {
-	    BLOCKING_REGION({
-		lock_func(th, mutex);
-	    }, lock_interrupt, mutex);
+	    int interrupted;
+	    int prev_status = th->status;
+	    int last_thread = 0;
+	    struct rb_unblock_callback oldubf;
 
-	    RUBY_VM_CHECK_INTS();
+	    set_unblock_function(th, lock_interrupt, mutex, &oldubf);
+	    th->status = THREAD_STOPPED_FOREVER;
+	    th->vm->sleeper++;
+	    th->locking_mutex = self;
+	    if (vm_living_thread_num(th->vm) == th->vm->sleeper) {
+		last_thread = 1;
+	    }
+
+	    th->transition_for_lock = 1;
+	    BLOCKING_REGION_CORE({
+		interrupted = lock_func(th, mutex, last_thread);
+	    });
+	    th->transition_for_lock = 0;
+	    remove_signal_thread_list(th);
+	    reset_unblock_function(th, &oldubf);
+
+	    th->locking_mutex = Qfalse;
+	    if (mutex->th && interrupted == 2) {
+		rb_check_deadlock(th->vm);
+	    }
+	    if (th->status == THREAD_STOPPED_FOREVER) {
+		th->status = prev_status;
+	    }
+	    th->vm->sleeper--;
+
+	    if (mutex->th == th) mutex_locked(th, self);
+
+	    if (interrupted) {
+		RUBY_VM_CHECK_INTS();
+	    }
 	}
     }
     return self;
 }
 
-/*
- * call-seq:
- *    mutex.unlock    => self
- *
- * Releases the lock.
- * Raises +ThreadError+ if +mutex+ wasn't locked by the current thread.
- */
-VALUE
-rb_mutex_unlock(VALUE self)
+static const char *
+mutex_unlock(mutex_t *mutex)
 {
-    mutex_t *mutex;
-    char *err = NULL;
-    GetMutexPtr(self, mutex);
+    const char *err = NULL;
+    rb_thread_t *th = GET_THREAD();
+    mutex_t *th_mutex;
 
     native_mutex_lock(&mutex->lock);
 
@@ -2394,16 +2614,85 @@ rb_mutex_unlock(VALUE self)
 	    /* waiting thread */
 	    native_cond_signal(&mutex->cond);
 	    mutex->cond_waiting--;
+	    mutex->cond_notified++;
 	}
     }
 
     native_mutex_unlock(&mutex->lock);
 
+    if (!err) {
+	GetMutexPtr(th->keeping_mutexes, th_mutex);
+	if (th_mutex == mutex) {
+	    th->keeping_mutexes = mutex->next_mutex;
+	}
+	else {
+	    while (1) {
+		mutex_t *tmp_mutex;
+		GetMutexPtr(th_mutex->next_mutex, tmp_mutex);
+		if (tmp_mutex == mutex) {
+		    th_mutex->next_mutex = tmp_mutex->next_mutex;
+		    break;
+		}
+		th_mutex = tmp_mutex;
+	    }
+	}
+	mutex->next_mutex = Qfalse;
+    }
+
+    return err;
+}
+
+/*
+ * call-seq:
+ *    mutex.unlock    => self
+ *
+ * Releases the lock.
+ * Raises +ThreadError+ if +mutex+ wasn't locked by the current thread.
+ */
+VALUE
+rb_mutex_unlock(VALUE self)
+{
+    const char *err;
+    mutex_t *mutex;
+    GetMutexPtr(self, mutex);
+
+    err = mutex_unlock(mutex);
     if (err) rb_raise(rb_eThreadError, err);
 
     return self;
 }
 
+static void
+rb_mutex_unlock_all(VALUE mutexes)
+{
+    const char *err;
+    mutex_t *mutex;
+
+    while (mutexes) {
+	GetMutexPtr(mutexes, mutex);
+	/* rb_warn("mutex #<%s:%p> remains to be locked by terminated thread",
+		rb_obj_classname(mutexes), (void*)mutexes); */
+	mutexes = mutex->next_mutex;
+	err = mutex_unlock(mutex);
+	if (err) rb_bug("invalid keeping_mutexes: %s", err);
+    }
+}
+
+static VALUE
+rb_mutex_sleep_forever(VALUE time)
+{
+    rb_thread_sleep_deadly();
+    return Qnil;
+}
+
+static VALUE
+rb_mutex_wait_for(VALUE time)
+{
+    const struct timeval *t = (struct timeval *)time;
+    rb_thread_wait_for(*t);
+    return Qnil;
+}
+
 VALUE
 rb_mutex_sleep(VALUE self, VALUE timeout)
 {
@@ -2416,19 +2705,18 @@ rb_mutex_sleep(VALUE self, VALUE timeout)
     rb_mutex_unlock(self);
     beg = time(0);
     if (NIL_P(timeout)) {
-	rb_thread_sleep_forever();
+	rb_ensure(rb_mutex_sleep_forever, Qnil, rb_mutex_lock, self);
     }
     else {
-	rb_thread_wait_for(t);
+	rb_ensure(rb_mutex_wait_for, (VALUE)&t, rb_mutex_lock, self);
     }
-    rb_mutex_lock(self);
     end = time(0) - beg;
     return INT2FIX(end);
 }
 
 /*
  * call-seq:
- *    mutex.sleep(timeout = nil)    => self
+ *    mutex.sleep(timeout = nil)    => number
  *
  * Releases the lock and sleeps +timeout+ seconds if it is given and
  * non-nil or forever.  Raises +ThreadError+ if +mutex+ wasn't locked by
@@ -2540,8 +2828,7 @@ barrier_alloc(VALUE klass)
     VALUE volatile obj;
     rb_barrier_t *barrier;
 
-    obj = Data_Make_Struct(klass, rb_barrier_t,
-			   barrier_mark, barrier_free, barrier);
+    obj = Data_Make_Struct(klass, rb_barrier_t, barrier_mark, barrier_free, barrier);
     barrier->owner = GET_THREAD();
     barrier->waiting = 0;
     barrier->tail = &barrier->waiting;
@@ -2689,7 +2976,7 @@ rb_exec_recursive(VALUE (*func) (VALUE, VALUE, int), VALUE obj, VALUE arg)
 /* tracer */
 
 static rb_event_hook_t *
-alloc_event_fook(rb_event_hook_func_t func, rb_event_flag_t events, VALUE data)
+alloc_event_hook(rb_event_hook_func_t func, rb_event_flag_t events, VALUE data)
 {
     rb_event_hook_t *hook = ALLOC(rb_event_hook_t);
     hook->func = func;
@@ -2714,7 +3001,7 @@ void
 rb_thread_add_event_hook(rb_thread_t *th,
 			 rb_event_hook_func_t func, rb_event_flag_t events, VALUE data)
 {
-    rb_event_hook_t *hook = alloc_event_fook(func, events, data);
+    rb_event_hook_t *hook = alloc_event_hook(func, events, data);
     hook->next = th->event_hooks;
     th->event_hooks = hook;
     thread_reset_event_flags(th);
@@ -2745,7 +3032,7 @@ set_threads_event_flags(int flag)
 void
 rb_add_event_hook(rb_event_hook_func_t func, rb_event_flag_t events, VALUE data)
 {
-    rb_event_hook_t *hook = alloc_event_fook(func, events, data);
+    rb_event_hook_t *hook = alloc_event_hook(func, events, data);
     rb_vm_t *vm = GET_VM();
 
     hook->next = vm->event_hooks;
@@ -2904,14 +3191,14 @@ thread_set_trace_func_m(VALUE obj, VALUE trace)
     GetThreadPtr(obj, th);
     rb_thread_remove_event_hook(th, call_trace_func);
 
-    if (!NIL_P(trace)) {
+    if (NIL_P(trace)) {
 	return Qnil;
     }
     thread_add_trace_func(th, trace);
     return trace;
 }
 
-static char *
+static const char *
 get_event_name(rb_event_flag_t event)
 {
     switch (event) {
@@ -2952,6 +3239,7 @@ call_trace_proc(VALUE args, int tracing)
     struct call_trace_func_args *p = (struct call_trace_func_args *)args;
     VALUE eventname = rb_str_new2(get_event_name(p->event));
     VALUE filename = rb_str_new2(rb_sourcefile());
+    VALUE argv[6];
     int line = rb_sourceline();
     ID id = 0;
     VALUE klass = 0;
@@ -2974,11 +3262,15 @@ call_trace_proc(VALUE args, int tracing)
 	    klass = rb_iv_get(klass, "__attached__");
 	}
     }
-    return rb_proc_call(p->proc, rb_ary_new3(6,
-					     eventname, filename, INT2FIX(line),
-					     id ? ID2SYM(id) : Qnil,
-					     p->self ? rb_binding_new() : Qnil,
-					     klass ? klass : Qnil));
+
+    argv[0] = eventname;
+    argv[1] = filename;
+    argv[2] = INT2FIX(line);
+    argv[3] = id ? ID2SYM(id) : Qnil;
+    argv[4] = p->self ? rb_binding_new() : Qnil;
+    argv[5] = klass ? klass : Qnil;
+    
+    return rb_proc_call_with_block(p->proc, 6, argv, Qnil);
 }
 
 static void
@@ -3040,6 +3332,8 @@ ruby_suppress_tracing(VALUE (*func)(VALUE, int), VALUE arg, int always)
 void
 Init_Thread(void)
 {
+#undef rb_intern
+
     VALUE cThGroup;
 
     rb_define_singleton_method(rb_cThread, "new", thread_s_new, -1);
@@ -3120,7 +3414,7 @@ Init_Thread(void)
 	/* main thread setting */
 	{
 	    /* acquire global interpreter lock */
-	    rb_thread_lock_t *lp = &GET_THREAD()->vm->global_interpreter_lock;
+	    rb_thread_lock_t *lp = &GET_THREAD()->vm->global_vm_lock;
 	    native_mutex_initialize(lp);
 	    native_mutex_lock(lp);
 	    native_mutex_initialize(&GET_THREAD()->interrupt_lock);
@@ -3128,13 +3422,84 @@ Init_Thread(void)
     }
 
     rb_thread_create_timer_thread();
+
+    (void)native_mutex_trylock;
+    (void)ruby_thread_set_native;
 }
 
 int
 ruby_native_thread_p(void)
 {
-    rb_thread_t *rb_thread_check_ptr(rb_thread_t *ptr);
     rb_thread_t *th = ruby_thread_from_native();
 
     return th ? Qtrue : Qfalse;
 }
+
+static int
+check_deadlock_i(st_data_t key, st_data_t val, int *found)
+{
+    VALUE thval = key;
+    rb_thread_t *th;
+    GetThreadPtr(thval, th);
+
+    if (th->status != THREAD_STOPPED_FOREVER || RUBY_VM_INTERRUPTED(th) || th->transition_for_lock) {
+	*found = 1;
+    }
+    else if (th->locking_mutex) {
+	mutex_t *mutex;
+	GetMutexPtr(th->locking_mutex, mutex);
+
+	native_mutex_lock(&mutex->lock);
+	if (mutex->th == th || (!mutex->th && mutex->cond_notified)) {
+	    *found = 1;
+	}
+	native_mutex_unlock(&mutex->lock);
+    }
+
+    return (*found) ? ST_STOP : ST_CONTINUE;
+}
+
+#if 0 /* for debug */
+static int
+debug_i(st_data_t key, st_data_t val, int *found)
+{
+    VALUE thval = key;
+    rb_thread_t *th;
+    GetThreadPtr(thval, th);
+
+    printf("th:%p %d %d %d", th, th->status, th->interrupt_flag, th->transition_for_lock);
+    if (th->locking_mutex) {
+	mutex_t *mutex;
+	GetMutexPtr(th->locking_mutex, mutex);
+
+	native_mutex_lock(&mutex->lock);
+	printf(" %p %d\n", mutex->th, mutex->cond_notified);
+	native_mutex_unlock(&mutex->lock);
+    }
+    else puts("");
+
+    return ST_CONTINUE;
+}
+#endif
+
+static void
+rb_check_deadlock(rb_vm_t *vm)
+{
+    int found = 0;
+
+    if (vm_living_thread_num(vm) > vm->sleeper) return;
+    if (vm_living_thread_num(vm) < vm->sleeper) rb_bug("sleeper must not be more than vm_living_thread_num(vm)");
+
+    st_foreach(vm->living_threads, check_deadlock_i, (st_data_t)&found);
+
+    if (!found) {
+	VALUE argv[2];
+	argv[0] = rb_eFatal;
+	argv[1] = rb_str_new2("deadlock detected");
+#if 0 /* for debug */
+	printf("%d %d %p %p\n", vm->living_threads->num_entries, vm->sleeper, GET_THREAD(), vm->main_thread);
+	st_foreach(vm->living_threads, debug_i, (st_data_t)0);
+#endif
+	rb_thread_raise(2, argv, vm->main_thread);
+    }
+}
